
/***********************************************************************

        Copyright (C) 1991,
        Virginia Polytechnic Institute & State University

        This program was originally written by Mr. Hyung K. Lee
        under the supervision of Dr. Dong S. Ha, in the Bradley
        Department of Electrical Engineering, VPI&SU, in 1991.

        This program is released for research use only. This program,
        or any derivative thereof, may not be reproduced nor used
        for any commercial product without the written permission
        of the authors.

        For detailed information, please contact to

        Dr. Dong S. Ha
        Bradley Department of Electrical Engineering
        Virginia Polytechnic Institute & State University
        Blacksburg, VA 24061

        Ph.: (540) 231-4942
        Fax: (540) 231-3362
        E-Mail: ha@vt.edu
        Web: http://www.ee.vt.edu/ha

        REFERENCE:
           H. K. Lee and D. S. Ha, "On the Generation of Test Patterns
           for Combinational Circuits," Technical Report No. 12_93,
           Dep't of Electrical Eng., Virginia Polytechnic Institute
           and State University.

***********************************************************************/

/**************************** HISTORY **********************************
 
        atalanta: version 1.0        H. K. Lee, 8/15/1991
        atalanta: version 1.1        H. K. Lee, 10/5/1992
        atalanta: version 2.0        H. K. Lee, 6/30/1997
 
***********************************************************************/

/*----------------------------------------------------------------- 
	filename structure.c

	list all modifications below:
	original:  10/5/1992	Hyung K. Lee   compiled and checked
-------------------------------------------------------------------*/
#include <stdio.h>

#include "parameter.h" 
#include "define.h"    
#include "macro.h"

extern int nog,nopi,nopo,noff,maxlevel,POlevel,PPOlevel,maxfout;
extern int *primaryin,*primaryout,*flip_flops;
extern GATEPTR *net;
extern STACKTYPE stack1,stack2,*event_list;
extern char *strcat(char *, const char *),*strcpy(char *, const char *);
extern HASHPTR symbol_tbl[];
extern HASHPTR FindHash(HASHPTR *table, int size, char *symbol, int key), InsertHash(HASHPTR *table, int size, char *symbol, int key);
extern void fatalerror(int errorcode);

#ifdef INCLUDE_HOPE
void allocate_stacks(void)
{
   ALLOCATE(stack1.list,GATEPTR,nog+SPAREGATES+nopo);
   ALLOCATE(stack2.list,GATEPTR,nog+SPAREGATES+nopo);
   clear(stack1);
   clear(stack2);
}

/*------compute_level----------------------------------------------
	computes psuedo-levels of each gate in net structure
	inputs : none
	outputs: returns real depth of the circuit, maxlevel+1

	Remark: PI, PPI --- level 0
		PO --- maxlevel+1 (if a PO gate is added)
		PPO --- maxlevel+2 (actual flip-flops)
-------------------------------------------------------------------*/
int compute_level(void)
{
   register i,j,flag=1;
   GATEPTR cg,ng;

   for(i=0;i<nog;i++) {
      cg=net[i];
      if(cg->fn==PI || cg->fn==DFF) {
	 cg->dpi=0;
	 push(stack1,net[i]);
	 cg->changed=cg->ninput;
      } else { cg->dpi=(-1); cg->changed=0; }
   }

   while(TRUE) {
      if(flag==1) while(!is_empty(stack1)) {
         cg=pop(stack1);
	 for(i=0;i<cg->noutput;i++) {
	    ng=cg->outlis[i];
	    if(++ng->changed==ng->ninput) {
	       ng->dpi=cg->dpi+1;
	       push(stack2,ng);
	    }
	 }
      } else if(flag==2) while(!is_empty(stack2)) {
         cg=pop(stack2);
	 for(i=0;i<cg->noutput;i++) {
	    ng=cg->outlis[i];
	    if(++ng->changed==ng->ninput) {
	       ng->dpi=cg->dpi+1;
	       push(stack1,ng);
	    }
	 }
      }
      flag = (flag==1) ? 2 : 1;
      if(is_empty(stack1) && is_empty(stack2)) break; /* exit */
   }

   /* Compute maxlevel */
   maxlevel=(-1);
   for(i=0;i<nopo;i++) {
      cg=net[primaryout[i]];
      if(cg->fn==PO) {
	 if(cg->dpi>maxlevel) { maxlevel=cg->dpi; flag=1; }
      } else if(cg->dpi>=maxlevel) { maxlevel=cg->dpi; flag=2; }
   }
   for(i=0;i<noff;i++) {
      cg=net[flip_flops[i]];
      for(j=0; j<cg->ninput; j++)
         if(cg->inlis[j]->dpi>=maxlevel) { maxlevel=cg->inlis[j]->dpi; flag=2; }
   }

   /* Renumber levels of POs and PPO(DFF)s */
   if(flag==1) maxlevel--;
   POlevel=maxlevel+1;
   PPOlevel=maxlevel+2;
   for(i=0;i<nopo;i++)
      if(net[primaryout[i]]->fn==PO) net[primaryout[i]]->dpi=POlevel;
   for(i=0;i<noff;i++) net[flip_flops[i]]->dpi=PPOlevel;
   
   return(maxlevel+1);
}

/*------allocate_event_list-------------------------------------------
	allocates levelized queue for event_list
	inputs : none
	outputs: event_list
        dpi of each gate and maxlevel should be set 
--------------------------------------------------------------------*/
void allocate_event_list(void)
{
    
   register int i;

   ALLOCATE(event_list,STACKTYPE,maxlevel+2);
   for(i=0;i<maxlevel+2;i++) clear(event_list[i]);
   for(i=0;i<nog;i++) event_list[net[i]->dpi].last++;
   for(i=0;i<maxlevel+2;i++) {
      ALLOCATE(event_list[i].list,GATEPTR,event_list[i].last+1+SIZE_OF_FUT);
      clear(event_list[i]);
   }
}

/*------levelize------------------------------------------------------
	gates are re-numbered and sorted in the levelized order
	affected data structure: 
		net,primaryin,flip_flops,event_list,symbol_tbl
--------------------------------------------------------------------*/
int levelize(void)
{
   register i,j,new=0;
   GATEPTR cg;

   /* re-number gates */
   for(i=0;i<nog;i++) pushgate(net[i]);
   for(i=0;i<maxlevel+2;i++) {
      for(j=0;j<=event_list[i].last;j++) {
         cg=event_list[i].list[j];
	 cg->index=new++;
      }
      clear(event_list[i]);
   }

   /* update gate numbers */
   for(i=0;i<nopi;i++)				/* primaryin */
      primaryin[i]=net[primaryin[i]]->index;
   for(i=0;i<nopo;i++)				/* primaryout */
      primaryout[i]=net[primaryout[i]]->index;
   for(i=0;i<noff;i++)				/* flip_flops */
      flip_flops[i]=net[flip_flops[i]]->index;

/*
   for(i=0;i<nog;i++) {
      cg=net[i];
      cg->symbol->index=cg->index;
   }
*/

   /* sort gates by index */
   i=0;
   while(i<nog) {
      if(i==net[i]->index) i++;
      else {			/* swap net[i] and net[net[i]->gid] */
	 cg=net[i];
	 net[i]=net[cg->index];
	 net[cg->index]=cg;
      }
   }
}

#else
/*------levelize: 
	topologically sorts the circuit in the levelized order------*/
int levelize(net, n, nopi, nopo, noff, stack)
GATEPTR *net; int n, nopi, nopo, noff; GATEPTR *stack;
{
   register int i, int_nog;
   register GATEPTR *first, *last;
   register GATEPTR ele, next;
   int G_index=0;
   int maxdpi=(-1);

   /* initialize */
   for(i=0; i<n; i++)
      if((ele=net[i]) != NULL) {
	 ele->changed=ele->ninput;
	 ele->dpi=(-1);
      }

   first=last=stack;	/* empty stack */

   /* Find gates with indegree=0 */
   for(i=0; i<nopi; i++) {
      ele=net[primaryin[i]];
      ele->changed=0;
      *last++=ele;
      primaryin[i]=G_index++;
   }
   for(i=0; i<noff; i++) {
      ele=net[flip_flops[i]];
      ele->changed=0;
      *last++=ele;
      flip_flops[i]=G_index++;
   }

   /* levelize */
   G_index=0;
   int_nog=0;
   while(first != last) {
      int_nog++;
      ele=*first++;
      ele->index=G_index++;
      if( ++(ele->dpi) > maxdpi) maxdpi=ele->dpi;
      for(i=0; i<ele->noutput; i++)
	 if((next=ele->outlis[i]) != NULL)
	    if(next->changed > 0) {
	       if(--(next->changed) == 0) *last++=next;
	       next->dpi = ele->dpi;
	    }
   }

   /* check for levelization */
   if(int_nog != n) {
      fprintf(stderr,"Error in circuit file.\n");
      fprintf(stderr,"Some gates are not reachable from PIs.\n");
      for(i=0; i<n; i++) {
         ele=net[i];
         if(ele->changed>0 || ele->dpi<0) {
            fprintf(stderr,"*** Unreachable gate from an input:");
	    fprintf(stderr,"%d'th element %s\n",ele->index,
	            ele->symbol->symbol);
         }
      }
      return(-1);
   }

   for(i=0; i<nopo; i++)	/* primaryout */
      primaryout[i]=net[primaryout[i]]->index;

   /* sort gates by index */
   for(i=0; i<n; ) {
      if(i==net[i]->index) i++;
      else {			/* swap net[i] and net[net[i]->index] */
	 ele=net[i];
	 net[i]=net[ele->index];
	 net[ele->index]=ele;
      }
   }

   return(maxdpi+1);
}
#endif

/*------add_PO-------------------------------------------------------
	adds a PO gate at each primary output
	should be called before levelize()

	modified by Hyung K Lee  2/15/1991
--------------------------------------------------------------------*/
int add_PO(void)
{
   register int i,j;
   GATEPTR gut,last,*outlist;
   char name[MAXSTRING];

   for(i=0;i<nopo;i++) {
      gut=net[primaryout[i]];
      if((last=net[nog])==NULL) ALLOCATE(last,GATETYPE,1);
      last->index=nog;
      last->fn=PO;
      last->ninput=1;
      ALLOCATE(last->inlis,GATEPTR,1);
      last->inlis[0]=gut;
      last->noutput=0;
      last->outlis=NULL;
#ifdef LEARNFLG
      last->plearn=NULL;
#endif
      strcpy(name,gut->symbol->symbol);
      strcat(name,"_PO");
      while((last->symbol=FindHash(symbol_tbl,HASHSIZE,name,0))
	 != NULL) strcat(name,"_PO");
      if((last->symbol=InsertHash(symbol_tbl,HASHSIZE,name,0))
	 == NULL) fatalerror(HASHERROR);
      else last->symbol->pnode=last;

      outlist=gut->outlis;
      ALLOCATE(gut->outlis,GATEPTR,gut->noutput+1);
      for(j=0;j<gut->noutput;j++) gut->outlis[j]=outlist[j];
      gut->outlis[gut->noutput]=last;
      gut->noutput+=1;

      primaryout[i]=nog;
      net[nog++]=last;
/*      if(outlist!=NULL) FREE(outlist); */
   }

   return(nopo);
}

/*------add_sparegates-----------------------------------------------
	allocates memory spaces for the fault injection.
	Should be called after levelization.
	Do not assign symbolic names.
	Inputs: none
--------------------------------------------------------------------*/
void add_sparegates(void)
{
   register int i;
   register GATEPTR gut;

   /* add sparegates */
   for(i=0;i<SIZE_OF_FUT;i++) {
      ALLOCATE(gut,GATETYPE,1);			/* CONSTANT Gate */
      gut->index=nog+2*i;
      gut->fn=DUMMY;
      gut->ninput=0;
      gut->inlis=NULL;
      gut->noutput=1;
      ALLOCATE(gut->outlis,GATEPTR,1);
      gut->dpi=0;
      reset(gut->changed);
      gut->symbol=NULL;
      gut->Gid=0;
      gut->stem=0;
      net[gut->index]=gut;

      ALLOCATE(gut,GATETYPE,1);			/* 2-input AND, OR, XOR */
      gut->index=nog+2*i+1;
      gut->fn=DUMMY;
      gut->ninput=2;
      ALLOCATE(gut->inlis,GATEPTR,2);
      gut->inlis[0]=net[nog+2*i];
      gut->noutput=0;
      ALLOCATE(gut->outlis,GATEPTR,maxfout);
      gut->dpi=0;
      reset(gut->changed);
      gut->Gid=0;
      gut->stem=0;
      gut->symbol=NULL;
      net[gut->index]=gut;
   }

   /* add one memory space for output list of POs */
   for(i=0; i<nopo; i++) {
      gut=net[primaryout[i]];
      if(gut->noutput==0) {
	 ALLOCATE(gut->outlis,GATEPTR,1);
	 gut->outlis[0]=(GATEPTR)NULL;
      }
   }
}

